@@ -68,7 +68,8 @@ public NoteManager (string directory) :
 
 			Tomboy.ExitingEvent += OnExitingEvent;
 		}
-
+		
+		
 		// Create the TrieController. For overriding in test methods.
 		protected virtual TrieController CreateTrieController ()
 		{
@@ -114,12 +115,14 @@ public NoteManager (string directory) :
 		{
 			if (NoteRenamed != null)
 				NoteRenamed (note, old_title);
+			this.notes.Sort (new CompareDates ());
 		}
 
 		void OnNoteSave (Note note)
 		{
 			if (NoteSaved != null)
 				NoteSaved (note);
+			this.notes.Sort (new CompareDates ());
 		}
 
 		protected virtual void CreateStartNotes ()
@@ -168,12 +171,12 @@ public NoteManager (string directory) :
 			try {
 				Note start_note = Create (Catalog.GetString ("Start Here"),
 				                          start_note_content);
-				start_note.QueueSave (true);
+				start_note.QueueSave (ChangeType.ContentChanged);
 				Preferences.Set (Preferences.START_NOTE_URI, start_note.Uri);
 
 				Note links_note = Create (Catalog.GetString ("Using Links in Tomboy"),
 				                          links_note_content);
-				links_note.QueueSave (true);
+				links_note.QueueSave (ChangeType.ContentChanged);
 
 				start_note.Window.Show ();
 			} catch (Exception e) {
@@ -200,6 +203,8 @@ public NoteManager (string directory) :
 					            e.Message);
 				}
 			}
+			
+			notes.Sort (new CompareDates ());
 
 			// Update the trie so addins can access it, if they want.
 			trie_controller.Update ();
@@ -220,7 +225,7 @@ public NoteManager (string directory) :
 						note.Window.Show ();
 
 					note.IsOpenOnStartup = false;
-					note.QueueSave (false);
+					note.QueueSave (ChangeType.NoChange);
 				}
 			}
 
@@ -248,8 +253,11 @@ public NoteManager (string directory) :
 			}
 
 			Logger.Log ("Saving unsaved notes...");
-
-			foreach (Note note in notes) {
+			
+			// Use a copy of the notes to prevent bug #510442 (crash on exit
+			// when iterating the notes to save them.
+			List<Note> notesCopy = new List<Note> (notes);
+			foreach (Note note in notesCopy) {
 				// If the note is visible, it will be shown automatically on
 				// next startup
 				if (note.HasWindow && note.Window.Visible)
@@ -451,7 +459,7 @@ public NoteManager (string directory) :
 				Tag tag = TagManager.GetOrCreateSystemTag (TagManager.TemplateNoteSystemTag);
 				template_note.AddTag (tag);
 
-				template_note.QueueSave (true);
+				template_note.QueueSave (ChangeType.ContentChanged);
 			}
 			
 			return template_note;
@@ -466,7 +474,7 @@ public NoteManager (string directory) :
 			        "</note-content>";
 
 			return string.Format (base_xml,
-			                      title,
+			                      XmlEncoder.Encode (title),
 			                      Catalog.GetString ("Describe your new note here."));
 		}
 
@@ -513,7 +521,7 @@ public NoteManager (string directory) :
 			get {
 				// FIXME: Only sort on change by listening to
 				//        Note.Saved or Note.Buffer.Changed
-				notes.Sort (new CompareDates ());
+				//notes.Sort (new CompareDates ());
 				return notes;
 			}
 		}
